/******************************************************************************
 *
 * File			:	jkmsg.c
 *	
 *	Author		:	Jayakumar B
 *
 *	Created 	:	28th April 2003
 *
 * Description	:	This file contains implementation of Linux Message Queues
 *
*******************************************************************************/
#include <linux/jkmsg.h> 

/* Timer value for random number generator */
extern unsigned long volatile jiffies;
extern struct task_struct *current;
extern volatile struct timeval xtime;

/*
 * JK : Flag for Debug ON/OFF
 */
#define JK_DEBUG 1

/* Global Variable declaration */
DSPTR  global_ds_head;	

/* Local Function declarations */
static DSPTR _msgq_ds_search(int value,int type);
static DSPTR _msgq_ds_malloc(void);
static MSGPTR _msgq_msg_malloc(void);
static MSGPTR _msgq_type_search(MSGPTR pointer,int type,MSGPTR *prev);
static void _msgq_ds_free(DSPTR queue_ptr,int qtype);
static int  _msgq_extract_perm(int msg_flag);
static int  _msgq_read_permission(DSPTR ptr);
static int _msgq_write_permission(DSPTR ptr);
static int _msgq_assign_mode(int permission);
static void _msgq_insert(DSPTR *pptr,const void *userptr,int msgsize);
static int _msgq_retrive(void **usermsg,MSGPTR msg_ptr,int size,int msgflg);



/* FIXME : This function is for debugging purpose */
static void myprint(int msgqid);


/******************************************************************************
 *
 *	Function Name 		:	JKmsgget()
 *
 *  Arguments			:	key,msgflg
 *
 *  Return Value		:	Integer
 *
 *  Description			:	This function implements the standard msgget() 
 *  						system call.Given the key and flag, it returns a 
 *  						new queue id or existing queue id
 *
*******************************************************************************/
asmlinkage int sys_JKmsgget(key_t key,int msgflg)
{

	DSPTR 	header,new_queue;
	int 	entry_count;

	entry_count = 0;

	/* Search the first level of linked list for given key */
	header = _msgq_ds_search(key,JK_KEY_SEARCH);

	if(header == NULL)
	{
		 printk("Creating new entry\n");

		/* Allocate memory for the new entry */
		new_queue = _msgq_ds_malloc();	
		if(new_queue == NULL)
		{
			printk("Memory error\n");
			return -ENOMEM;
		} 

		/* Create a new entry in the first level */

		/* Check if the first entry itself has to be created */
		if(global_ds_head == NULL) {
			/* Initialize the entry */
			global_ds_head = new_queue;
			global_ds_head->msg_perm.key = key;
			global_ds_head->msg_perm.mode = _msgq_assign_mode(msgflg);
			return(global_ds_head->msg_perm.seq);
		} else {
			/* Traverse the linked list till the end */
			printk("Assigning the new : %d\n",entry_count);
			header = global_ds_head;
			while(header->ds_next != NULL) {
				entry_count++;
				header = header->ds_next;
			}

			/* Check for maximum number of entries in the list */
			if(entry_count == JK_MSGQ_MAX_ENTRY)
			{
				_msgq_ds_free(new_queue,JK_QUEUE_EMPTY);
				return -ENOSPC;
			} 

			/* Add the new entry and initialize the entry */
			header->ds_next = new_queue;
			new_queue->msg_perm.key = key;
			new_queue->msg_perm.mode = _msgq_assign_mode(msgflg);
			/* Return the message queue id */
			return(new_queue->msg_perm.seq);
		}
	} else {

		/* If the queue is already created then proceed based on the flags */
		entry_count  = _msgq_extract_perm(msgflg);

		printk("Getting the old entry\n");
		
		/*
		 * Check whether both JK_IPC_EXCL and JK_IPC_CREAT 
		 * is enabled or not.In that case if the queue already
		 * exists this system call fails 
		 * 
		 * */
		 
		if(entry_count  == JK_IPC_EXCL_CREAT)
			return -EEXIST;
  
		/* Check for the read permission of the queue */
		 if(_msgq_read_permission(header) == JK_FAILURE) 
			return -EACCES;
		
		/* Return the existing queue id */
		printk("Here it is :%d",header->msg_perm.seq);
		return(header->msg_perm.seq);

	}
}
/******************************************************************************
 *
 *	Function Name 		:	JKmsgsnd()
 *
 *  Arguments			:	msgqid,msgp,msgsz,msgflg
 *
 *  Return Value		:	Integer
 *
 *  Description			:	This function implements standard msgsnd() system
 *  						call.The sending process calls this system call 
 *  						with message and proper parameters.This enqueues
 *  						the message which is later dequeued by the receive
 *  						proces
 *
*******************************************************************************/
asmlinkage int sys_JKmsgsnd(int msgqid,const void *msgp,int msgsz,int msgflg)
{

	DSPTR ds_ptr;
	int	  msg_perm;
	struct JK_msgbuf *msg_ptr;


			
	/* Search the queue with given message queue id */
	ds_ptr  = _msgq_ds_search(msgqid,JK_QID_SEARCH);


	/* Check for error conditions */
	if(ds_ptr == NULL) 
		return -EIDRM;

	/* Check for the write permission into the queue */
	msg_perm = _msgq_write_permission(ds_ptr); 
	if(msg_perm == JK_FAILURE) 
		return -EACCES;

	msg_ptr = (struct JK_msgbuf *)msgp;
	/* Check the message buffer pointer */
	if(msg_ptr == NULL) 
		return - EFAULT;

	/* Check for other illegal values */
	if((msg_ptr->mtype < 0)||(msgqid < 0) || (msgsz < 0)) 
		return -EINVAL;

	/* Check whether the queue is full or not */
	if(ds_ptr->msg_qnum == JK_MSGQ_MAX_ENTRY)
	{
		/* Check wether to return immediately or not */
		if(msgflg == JK_IPC_NOWAIT) 
		{
	   		printk("IPC_NOWAIT_ENABLED ! Returning immediately\n");
       		return(JK_SUCCESS);
		}
		else
		{
			/* Wait until someone reads from this message queue */
			interruptible_sleep_on(ds_ptr->wwait);

			/* Some message vacant slot is available in the queue */
			_msgq_insert(&ds_ptr,msgp,msgsz);
		}
	}
	else  
		/* Insert in the second level */
#ifdef JK_DEBUG
	printk("\n DEBUG : Slots are remaining in the queue !! \n");
#endif
		_msgq_insert(&ds_ptr,msgp,msgsz);

	/* 
	 * Check whether the message is written into the queue for the first time
	 * and somebody is waiting to read from the queue
	 */

	// if(ds_ptr->msg_qnum == 0 && ds_ptr->rwait->task_list.next != NULL)
	if(ds_ptr->msg_qnum == 0)
#ifdef JK_DEBUG
	printk("\n DEBUG : Looks like somebody is waiting to read \n");
#endif
			wake_up_interruptible(ds_ptr->rwait);

#ifdef JK_DEBUG
	printk("\n DEBUG : After sending the message \n");
	myprint(msgqid);
#endif
	
	return(JK_SUCCESS);
}
/******************************************************************************
 *
 *	Function Name 		:
 *
 *  Arguments			:
 *
 *  Return Value		:
 *
 *  Description			:
 *
*******************************************************************************/
asmlinkage int sys_JKmsgrcv(int msgqid,const void *msgp,int msgsz,long mtype,int msgflg)
{
	
 	DSPTR 	msg_ptr;
	MSGPTR	ptr,prev_ptr;
	int	  	msg_perm,msg_size,offset_size;

#ifdef JK_DEBUG
	printk("\n DEBUG : Entering the system call for msgrecv\n");
#endif

#ifdef JK_DEBUG
	printk("\n DEBUG : The message queue contains.... \n");
	myprint(msgqid);
#endif
	
    /* Search for the message queue with the given queue id */
	msg_ptr  = _msgq_ds_search(msgqid,JK_QID_SEARCH);

#ifdef JK_DEBUG
	if(msg_ptr != NULL)
		printk("\n DEBUG : Message queue found\n");
	else
		printk("\n DEBUG : Message queue not found\n");
#endif

	prev_ptr = msg_ptr->msg_first;

	if(msg_ptr == NULL) 
		return - EIDRM;

	/* Check for the read permission */
	msg_perm = _msgq_read_permission(msg_ptr); 

#ifdef JK_DEBUG
	if(msg_perm == JK_SUCCESS)
		printk("\n DEBUG : Read permission is there\n");
	else
		printk("\n DEBUG : Read permission is not there\n");
#endif

	if(msg_perm == JK_FAILURE) 
		return -EACCES;

	if((msgqid < 0) || (mtype < 0)) 
		return -EINVAL;

	if(mtype != 0) {
		printk("Called from UP !!\n");
		/* Search for the given message type */
		ptr = _msgq_type_search(msg_ptr->msg_first,mtype,&prev_ptr);
	}
	else 
	{
		printk("Called from DOWN !!\n");
		ptr = msg_ptr->msg_first;
	}

	/* 
	 * If the message is not found and the wait is disabled then return 
	 * Otherwise wait still someone writes into the queue
	 */
	if(ptr == NULL && (msgflg & JK_IPC_NOWAIT == JK_IPC_NOWAIT)) 
		return -EIDRM;
	else
	{
		/*
		 * Sleep till somebody wakes up
		 */
#ifdef JK_DEBUG
	printk("\n DEBUG : Blocking Receive...Going to sleep\n");
#endif
		interruptible_sleep_on(msg_ptr->rwait);	
#ifdef JK_DEBUG
	printk("\n DEBUG : Somebody had written...Awake\n");
#endif
				
	ptr = msg_ptr->msg_first;

	if(ptr->msg_ts < msgsz) 
		return -EINVAL;
			
	/* Check for the JK_MSG_NOERROR flag */
#ifdef JK_DEBUG
	printk("\n DEBUG : Value is :%d \n",msgflg & JK_MSG_NOERROR);
	printk("\n DEBUG : Value is :%d \n",JK_MSG_NOERROR);
#endif


	if((msgflg & JK_MSG_NOERROR) == JK_MSG_NOERROR) {
		printk("The no error bit is set\n");
		memcpy(msgp,(void *)ptr->msg_spot,msgsz); 
		printk("The requested size :%d\n",msgsz);
		printk("The total size     :%d\n",ptr->msg_ts);
		offset_size = ptr->msg_ts-msgsz;
		printk("The offset size    :%d\n",offset_size);
 		ptr->msg_ts = offset_size;
		memcpy(ptr->msg_spot,(ptr->msg_spot)+msgsz,offset_size);
		
	} else {

#ifdef JK_DEBUG
	printk("\n DEBUG : JK_MSG_NOERROR is not set\n");
#endif

		/* Check for the normal condition */
		if(ptr->msg_ts == msgsz) {

#ifdef JK_DEBUG
	printk("\n DEBUG : Equal size of message\n");
#endif

			/* Copy the message contents as it is */
			memcpy(msgp,(void *)ptr->msg_spot,msgsz); 
	
			/* Free the read message */
			if(ptr == msg_ptr->msg_first)
			{
				prev_ptr = msg_ptr->msg_first;
				msg_ptr->msg_first = msg_ptr->msg_first->msg_next;
				kfree(prev_ptr);
			} else {
				/* Delete at the end */
				if(ptr->msg_next == NULL)
					prev_ptr->msg_next = NULL;
				 else 
					prev_ptr->msg_next = ptr->msg_next;
					
					kfree(ptr);
			}
			return(JK_SUCCESS);
		} else 
			/* Message size is less and JK_MSG_NOERROR is not set */
			/* Return proper error code */
			return -E2BIG;
	}
	}

	/* Update the statistics */
	msg_ptr->msg_qnum--;
	msg_ptr->msg_cbytes -= msgsz;
	msg_ptr->msg_lrpid   = current->pid;
	msg_ptr->msg_rtime   = xtime.tv_usec;

	/* 
	 * Check if somebody is waiting in the queue to write
	 * and wake them up if nessary
	 */

#if 0
	if(msg_ptr->wwait)
		wake_up_interruptible(&msg_ptr->wwait);	
#endif	

	return(JK_SUCCESS);
}				
/******************************************************************************
 *
 *	Function Name 		:
 *
 *  Arguments			:
 *
 *  Return Value		:
 *
 *  Description			:
 *
*******************************************************************************/
asmlinkage int sys_JKmsgctl(int msgqid,int cmd,DSPTR buf)
{

	DSPTR msg_ptr;
	int	  msg_perm;
	
	switch(cmd) {

	case JK_IPC_STAT :
		msg_ptr  = _msgq_ds_search(msgqid,JK_QID_SEARCH);
		if(msg_ptr == NULL) 
			return -EIDRM;
	
		/* Check for the read permission */	
		msg_perm = _msgq_read_permission(msg_ptr); 
	
		if(msg_perm == JK_FAILURE) 
			return -EACCES;
		
		/* Copy the JK_msgqid_ds to the buffer and return */
		memcpy(buf,msg_ptr,sizeof(struct JK_msgqid_ds));
		return(JK_SUCCESS);
		break;

	case JK_IPC_SET :
		msg_ptr  = _msgq_ds_search(msgqid,JK_QID_SEARCH);

		if(msg_ptr == NULL) 
			return -EIDRM;
	
		/* Check for the write permission */
		msg_perm = _msgq_write_permission(msg_ptr); 
	
		if(msg_perm == JK_FAILURE) 
			return -EACCES;
	
		/* Copy the ipc_perm values to the buffer */	
		memcpy(msg_ptr,buf,sizeof(struct JK_ipc_perm));
		return(JK_SUCCESS);
		break;
		
case JK_IPC_RMID :
 		msg_ptr  = _msgq_ds_search(msgqid,JK_QID_SEARCH);

		if(msg_ptr == NULL) 
			return -EIDRM;
	
		/* Check for the write permission */ 	
		msg_perm = _msgq_write_permission(msg_ptr); 
	
		if(msg_perm == JK_FAILURE) 
			return -EACCES;
	
		_msgq_ds_free(msg_ptr,JK_QUEUE_NOT_EMPTY);

		return(JK_SUCCESS);
	
		break;
	}	
	return(JK_SUCCESS);
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_write_pemission()
 *
 *  Arguments			:	ptr
 *
 *  Return Value		:	Integer
 *
 *  Description			:	This is a local function.Given the queue it checks
 *  						whether the calling process has got the write perm
 *  						in to the queue or not.Otherwise it returns the 
 *  						appropriate error codes.
 *
*******************************************************************************/
int _msgq_write_permission(DSPTR ptr)
{

	int ipc_mode;

	/* FIXME : Fix the permission */

	return(JK_SUCCESS);

	ipc_mode = ptr->msg_perm.mode;

	/* Check for all permissions */
	if(ipc_mode & JK_ALL_WRITE) 
		return(JK_SUCCESS);
				
	/* Check if the process id is equal to the group id */
	if(ipc_mode & JK_GROUP_WRITE) {
		if(ptr->msg_perm.gid == current->gid)
			return(JK_SUCCESS);
	}

	/* Check if the owner process itself wants to access the queue */
	if(ipc_mode & JK_USER_WRITE) {
		if(ptr->msg_perm.uid == current->pid)
				return(JK_SUCCESS);
	}

	return(JK_FAILURE);
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_read_permission()
 *
 *  Arguments			:	DSPTR
 *
 *  Return Value		:	Integer
 *
 *  Description			:	This is a local function.Given the queue it checks
 *  						whether the calling process has got the read perm
 *  						in to the queue or not.Otherwise it returns the 
 *  						appropriate error codes.
 *
*******************************************************************************/
int _msgq_read_permission(DSPTR ptr)
{

	int ipc_mode;

	ipc_mode = ptr->msg_perm.mode;

	/* FIXME : Fix the permission */
	return(JK_SUCCESS);

	/* Check for all permissions */
	if(ipc_mode & JK_ALL_READ) 
		return(JK_SUCCESS);
				
	/* Check if the process id is equal to the group id */
	if(ipc_mode & JK_GROUP_READ) {
		
		if(ptr->msg_perm.gid == current->gid)
			return(JK_SUCCESS);
	}

	/* Check if the owner process itself wants to access the queue */
	if(ipc_mode & JK_USER_READ) {
		if(ptr->msg_perm.uid == current->pid)
				return(JK_SUCCESS);
	}

	return(JK_FAILURE);
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_insert()
 *
 *  Arguments			:	pptr,userptr,msgsize
 *
 *  Return Value		:	Void
 *
 *  Description			:	This is a local function and inserts a new message
 *  						in the queue.After inserting appropriate pointers 
 *  						are updated.
 *
*******************************************************************************/
void _msgq_insert(DSPTR *pptr,const void *userptr,int msgsize)
{
	MSGPTR first,last,temp;
	struct JK_msgbuf *msgptr;
	

	msgptr = (struct JK_msgbuf *)userptr;



	/* Store the pointers in temperory variables */
	first = (*pptr)->msg_first;
	last  = (*pptr)->msg_last;

	/* Inserting in the beginning */
   if(first ==  NULL) 
   {
			first = _msgq_msg_malloc();
			/* Make the first point to appropriate location */
			first->msg_type = msgptr->mtype;
			first->msg_spot = (char*)kmalloc(msgsize,3);
			memcpy(first->msg_spot,(char *)userptr,msgsize);
			first->msg_ts   = msgsize;
			
			(*pptr)->msg_first = first;
			(*pptr)->msg_last  = first;


	} 
	else 
	{
			
			/* Insert at the end */
			while(first->msg_next != NULL)
				first = first->msg_next;

			/* Fill the structure with appropriate values */
			first->msg_next = _msgq_msg_malloc();
			first->msg_next->msg_type = msgptr->mtype;
			first->msg_next->msg_spot = (char*)kmalloc(msgsize,3);
			memcpy(first->msg_next->msg_spot,(char *)userptr,msgsize);
			first->msg_next->msg_ts   = msgsize;
		
			(*pptr)->msg_last = first->msg_next; 
     }

	/* Update the statistics and time */
	(*pptr)->msg_qnum++;
	(*pptr)->msg_cbytes += msgsize;
	(*pptr)->msg_lspid   = current->pid;
	(*pptr)->msg_stime   = xtime.tv_usec;
	

}
/******************************************************************************
 *
 *	Function Name 		:
 *
 *  Arguments			:
 *
 *  Return Value		:
 *
 *  Description			:
 *
*******************************************************************************/
int _msgq_retrive(void **usermsg,MSGPTR msg_ptr,int size,int msgflg)
{
	char *new_area;

	/* Check for the message flag */
	if((msgflg & JK_MSG_NOERROR) == JK_MSG_NOERROR) {
		/* Check for the message size */
		if(size < msg_ptr->msg_ts && size != 0)
		{
			/* Copy only number of bytes requested by the user and 
			 * Place the remaining message in the queue */
			printk("The mtype is :%d\n",msg_ptr->msg_type);
			memcpy(usermsg,msg_ptr->msg_spot,size);
			new_area = msg_ptr+size;
			kfree(msg_ptr);
			msg_ptr = new_area;
			return(JK_SUCCESS);
		} else 
			return(JK_FAILURE);
	} else {
		/* In case the msg flag is not set to JK_MSG_NOERROR 
		 * report error */
		printk("The mtype is **:%d\n",msg_ptr->msg_type);
		if(size < msg_ptr->msg_ts && size != 0)
			return(JK_FAILURE);
		else {
			printk("The size : %d\n",msg_ptr->msg_ts);
			memcpy(*usermsg,msg_ptr->msg_spot,sizeof(msg_ptr->msg_ts));
			return(JK_SUCCESS);
		}
	}
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_ds_search()
 *
 *  Arguments			:	value,type
 *
 *  Return Value		:	DSPTR
 *
 *  Description			:	This is a local function which searches for a given
 *  						message queue.
 *
*******************************************************************************/
DSPTR _msgq_ds_search(int value,int type)
{

	DSPTR	ds_head;
	/* Store the global head into the local variable */
	ds_head = global_ds_head;
	
	if(ds_head == NULL) {
		printk("Pointer is NULL\n");
		return(NULL);
	}

	/* Traverse the list at first level and get the pointer */
	while(ds_head != NULL) {
		/* Search for the appropriate field */
		/* Currently JK_QID_SEARCH and JK_KEY_SEARCH are supported */
		
		if(type == JK_QID_SEARCH) {
			if(ds_head->msg_perm.seq == value)
				break;
		} else {
			if(ds_head->msg_perm.key == value)
				break;
		}
		ds_head = ds_head->ds_next;
	}
	
	/* Return the pointer */
	return(ds_head);
}
/******************************************************************************
 *
 *	Function Name 		:
 *
 *  Arguments			:
 *
 *  Return Value		:
 *
 *  Description			:
 *
*******************************************************************************/
MSGPTR _msgq_type_search(MSGPTR pointer,int type,MSGPTR *prev)
{
	/* Traverse the list from the starting */
	while(pointer!=NULL)
	{
		/* Check for the message type */
		if(pointer->msg_type == type)
				return(pointer);

		pointer = pointer->msg_next;
		*prev = pointer;
	}
	return(NULL);
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_ds_malloc()
 *
 *  Arguments			:	None
 *
 *  Return Value		:	DSPTR
 *
 *  Description			:	This function allocates memory for the message 
 *  						queue data structure and does initialization of the
 *  						data structure
 *
*******************************************************************************/
DSPTR _msgq_ds_malloc()
{
	DSPTR ds_ptr;
	int   msg_id;
	pid_t ppid;
	struct task_struct *current_ptr;


	ds_ptr = (DSPTR)kmalloc(sizeof(struct JK_msgqid_ds),3);
	if(ds_ptr == NULL)
		return(NULL);
	else {
		/* Do some basic initialization of pointers */
		ds_ptr->msg_first = NULL;
		ds_ptr->msg_last  = NULL;
		ds_ptr->ds_next   = NULL;
		
		/* Allocate memory for read and write wait queues */
		ds_ptr->rwait = (struct __wait_queue_head *)kmalloc(sizeof(struct __wait_queue_head),3);

		ds_ptr->wwait     = (struct __wait_queue_head *)kmalloc(sizeof(struct __wait_queue_head),3);
	
		/* Do some basic initialization of data */
		ds_ptr->msg_qnum    = 0;
		ds_ptr->msg_cbytes = 0;

		/* Assign the process id of the current process */
		ds_ptr->msg_perm.uid  = current->pid;
		ds_ptr->msg_perm.gid  = current->gid;
		ds_ptr->msg_perm.cuid = current->uid;
		ds_ptr->msg_perm.cgid = current->gid;

		/* Assign the maximum number of bytes */
		ds_ptr->msg_qbytes = JK_MAX_MSG_BYTES;

		/* Set the message queue id to some random integer */
		msg_id  = jiffies; 
		ds_ptr->msg_perm.seq = msg_id%65535; 

		return(ds_ptr);
	}
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_ds_free()
 *
 *  Arguments			:	DSPTR,qtype
 *
 *  Return Value		:	Void
 *
 *  Description			:	This is a local function written for freeing the 
 *  						message queue data structure.
 *
*******************************************************************************/
void _msgq_ds_free(DSPTR queue_ptr,int qtype)
{

	MSGPTR msg_ptr,temp;

	/* Check for a new entry */
	if(qtype == JK_QUEUE_EMPTY)		
		kfree(queue_ptr);
	else {

		/* 
		 * Queue is already having some messages in it 
		 * While deleting the queue first delete the messages and then
		 * delete the message first level data structure 
		 * */
			
		if(queue_ptr == global_ds_head) 
			global_ds_head = global_ds_head->ds_next;

		msg_ptr = queue_ptr->msg_first;
		while(msg_ptr != NULL) {
			temp = msg_ptr->msg_next;
			kfree(msg_ptr);
			msg_ptr = temp;
		}

		kfree(queue_ptr);
	}
		
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_msg_malloc()
 *
 *  Arguments			:	None
 *
 *  Return Value		:	MSGPTR
 *
 *  Description			:	This function allocates memory for the message 
 *  						queue data structure and does initialization of the
 *  						data structure
 *
*******************************************************************************/
MSGPTR _msgq_msg_malloc()
{
	MSGPTR msgptr;

	msgptr = (MSGPTR)kmalloc(sizeof(struct JK_msg),3);
	msgptr->msg_next = NULL;
	msgptr->msg_spot = NULL;

	return(msgptr);
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_extract_perm()
 *
 *  Arguments			:	msg_flag
 *
 *  Return Value		:	Integer
 *
 *  Description			:	This is a local function which extracts permission
 *  						fields.
 *
*******************************************************************************/
int  _msgq_extract_perm(int msg_flag)
{
	/* Check whether both JK_IPC_CREAT and JK_IPC_EXCL is enabled */
	if((msg_flag & JK_IPC_EXCL_CREAT) == JK_IPC_EXCL_CREAT)
			return(JK_IPC_EXCL_CREAT);

	/* Check whether JK_IPC_CREAT and JK_IPC_EXCL is enabled */
	if((msg_flag & JK_IPC_CREAT) == JK_IPC_CREAT)
			return(JK_IPC_CREAT);
			
	/* Check whether both JK_IPC_CREAT and JK_IPC_EXCL is enabled */
	if((msg_flag & JK_IPC_EXCL) == JK_IPC_EXCL)
			return(JK_IPC_EXCL);
}
/******************************************************************************
 *
 *	Function Name 		:	_msgq_assign_mode()
 *
 *  Arguments			:	permission
 *
 *  Return Value		:	Integer
 *
 *  Description			:	This is a local function.It assigns the mode.
 *
******************************************************************************/
int _msgq_assign_mode(int permission)
{
	int msg_perm;

	/* Extract the parameter */
	msg_perm = permission & JK_IPC_PERM_FIELD;
	return(msg_perm);
}
/******************************************************************************
 *
 *	Function Name 		:
 *
 *  Arguments			:
 *
 *  Return Value		:
 *
 *  Description			:
 *
*******************************************************************************/
void myprint(int msgqid)
{
	
	DSPTR  ds_ptr;
	MSGPTR msg_ptr;
	struct mymsg *mptr;

	/* FIXME : This function is for debug purpose.
	 * Remove it after debugging
	 * */

	ds_ptr  = _msgq_ds_search(msgqid,JK_QID_SEARCH);
	msg_ptr = ds_ptr->msg_first;

	printk("Here is the message list\n");

	while(msg_ptr) {
		
		printk("The mtype is : %d\n",msg_ptr->msg_type);
		mptr = (struct mymsg *)msg_ptr->msg_spot;
		printk("The contents are : %d..%d..%d\n",mptr->mtype,mptr->tos,mptr->length);
		msg_ptr = msg_ptr->msg_next;
	}
}

